﻿using Grpc.Net.Client;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Timers;

namespace Overt.Core.Grpc.H2
{
    internal class EndpointStrategy : IEndpointStrategy
    {
        #region Constructor
        private readonly object _lock = new object();
        private readonly Timer _timer;
        private readonly ConcurrentDictionary<string, IEndpointDiscovery> _discoveries = new ConcurrentDictionary<string, IEndpointDiscovery>();
        private readonly ConcurrentDictionary<string, List<ChannelWrapper>> _channelWrappers = new ConcurrentDictionary<string, List<ChannelWrapper>>();
        private readonly ConcurrentDictionary<string, List<string>> _targets = new ConcurrentDictionary<string, List<string>>();

        EndpointStrategy()
        {
            _timer = new Timer(ClientTimespan.ResetInterval.TotalSeconds * 1000);
            InitCheckTimer();
        }
        #endregion

        #region Destructor
        ~EndpointStrategy()
        {
            _timer?.Stop();
            _timer?.Dispose();

            foreach (var item in _channelWrappers)
            {
                item.Value?.ForEach(channel =>
                {
                    channel.ShutdownAsync();
                });
            }
            _channelWrappers.Clear();
        }
        #endregion

        #region Instance
        private readonly static object _instanceLocker = new object();
        private static EndpointStrategy _endpintStrategy;
        public static EndpointStrategy Instance
        {
            get
            {
                if (_endpintStrategy != null)
                    return _endpintStrategy;
                lock (_instanceLocker)
                {
                    if (_endpintStrategy != null)
                        return _endpintStrategy;

                    _endpintStrategy = new EndpointStrategy();
                    return _endpintStrategy;
                }
            }
        }
        #endregion

        #region Public Property
        /// <summary>
        /// 节点有变动
        /// </summary>
        public Action NodeChanged { get; set; }
        #endregion

        #region Public Method
        /// <summary>
        /// 添加ServiceDiscovery
        /// </summary>
        /// <param name="serviceDiscovery"></param>
        public void AddServiceDiscovery(IEndpointDiscovery serviceDiscovery)
        {
            if (serviceDiscovery == null)
                return;

            serviceDiscovery.Watched = () =>
            {
#if NET5_0_OR_GREATER
                GetSetTargets(serviceDiscovery.Options.ServiceName, false);
#else
                GetSetChannels(serviceDiscovery.Options.ServiceName, false);
#endif
                if (NodeChanged != null)
                    NodeChanged();
            };
            _discoveries.AddOrUpdate(serviceDiscovery.Options.ServiceName, serviceDiscovery, (k, v) => serviceDiscovery);
        }

        /// <summary>
        /// 获取所有
        /// </summary>
        /// <param name="serviceName"></param>
        /// <returns></returns>
        public List<ChannelWrapper> GetChannelWrappers(string serviceName)
        {
            if (_channelWrappers.TryGetValue(serviceName, out List<ChannelWrapper> channels) &&
                channels.Count > 0)
                return channels;

            lock (_lock)
            {
                if (_channelWrappers.TryGetValue(serviceName, out channels) &&
                    channels.Count > 0)
                    return channels;

                channels = GetSetChannels(serviceName);
                if ((channels?.Count ?? 0) <= 0 && ServiceBlackPolicy.Exist(serviceName))
                    channels = GetSetChannels(serviceName, false);

                return channels;
            }
        }

        /// <summary>
        /// 获取所有
        /// </summary>
        /// <param name="serviceName"></param>
        /// <returns></returns>
        public List<string> GetTargets(string serviceName)
        {
            if (_targets.TryGetValue(serviceName, out List<string> targets) &&
                targets.Count > 0)
                return targets;

            lock (_lock)
            {
                if (_targets.TryGetValue(serviceName, out targets) &&
                    targets.Count > 0)
                    return targets;

                targets = GetSetTargets(serviceName);
                if ((targets?.Count ?? 0) <= 0 && ServiceBlackPolicy.Exist(serviceName))
                    targets = GetSetTargets(serviceName, false);

                return targets;
            }
        }



        /// <summary>
        /// 获取callinvoker
        /// 随机获取
        /// </summary>
        /// <param name="serviceName"></param>
        /// <returns></returns>
        public ChannelWrapper GetChannelWrapper(string serviceName)
        {
            var channels = GetChannelWrappers(serviceName);
            return ServicePollingPolicy.Random(channels);
        }

        /// <summary>
        /// 屏蔽
        /// </summary>
        /// <param name="serviceName"></param>
        /// <param name="failedCallInvoker"></param>
        public void Revoke(string serviceName, ChannelWrapper channelWrapper)
        {
            lock (_lock)
            {
                if (channelWrapper == null)
                    return;

                // channels
                if (_channelWrappers.TryGetValue(serviceName, out List<ChannelWrapper> channels) &&
                    channels.Any(oo => oo == channelWrapper))
                {
                    channels.Remove(channelWrapper);
                    _channelWrappers.AddOrUpdate(serviceName, channels, (k, v) => channels);
                }

                // add black
                ServiceBlackPolicy.Add(serviceName, channelWrapper.Target);

                channelWrapper.ShutdownAsync();

                // reinit callinvoker
                if (channels.Count <= 0)
                    GetSetChannels(serviceName, false);
            }
        }

        /// <summary>
        /// 初始化Timer
        /// </summary>
        public void InitCheckTimer()
        {
            _timer.Elapsed += (sender, e) =>
            {
                lock (_lock)
                {
                    _timer.Stop();

                    try
                    {
                        foreach (var item in _channelWrappers)
                        {
                            GetSetChannels(item.Key);
                        }
                    }
                    catch { }

                    _timer.Start();
                }
            };
            _timer.Start();
        }
#endregion

#region Private Method
        /// <summary>
        /// 初始化callinvoker
        /// </summary>
        /// <param name="serviceName"></param>
        /// <param name="filterBlack">过滤黑名单 default true</param>
        /// <returns></returns>
        private List<ChannelWrapper> GetSetChannels(string serviceName, bool filterBlack = true)
        {
            if (!_discoveries.TryGetValue(serviceName, out IEndpointDiscovery discovery))
                return null;

            _channelWrappers.TryGetValue(serviceName, out List<ChannelWrapper> channelWrappers);
            channelWrappers ??= new List<ChannelWrapper>();
            var targets = discovery.FindServiceEndpoints(filterBlack);
            if ((targets?.Count ?? 0) <= 0)
            {
                // 如果consul 取不到 暂时直接使用本地缓存的连接（注册中心数据清空的情况--异常）
                _channelWrappers.TryGetValue(serviceName, out channelWrappers);
                return channelWrappers;
            }

            foreach (var target in targets)
            {
                if (channelWrappers.Any(oo => oo.Target == target.target))
                    continue;

                var scheme = "http";
                if (!string.IsNullOrWhiteSpace(discovery.Options.Scheme))
                    scheme = discovery.Options.Scheme;
                var channel = GrpcChannel.ForAddress($"{scheme}://{target.target}", discovery.Options.GrpcChannelOptions);
                var channelWrapper = new ChannelWrapper(target.serviceId, channel);
                channelWrappers.Add(channelWrapper);
            }

            // 移除已经销毁的callInvokers
            var destroyChannels = channelWrappers.Where(oo => !targets.Any(target => target.target == oo.Target)).ToList();
            foreach (var channel in destroyChannels)
            {
                channelWrappers.Remove(channel);
                channel.ShutdownAsync();
            }

            _channelWrappers.AddOrUpdate(serviceName, channelWrappers, (key, value) => channelWrappers);

            var newTargets = channelWrappers.Select(oo => oo.Target).ToList();
            _targets.AddOrUpdate(serviceName, newTargets, (key, value) => newTargets);
            return channelWrappers;
        }

        /// <summary>
        /// 初始化callinvoker
        /// </summary>
        /// <param name="serviceName"></param>
        /// <param name="filterBlack">过滤黑名单 default true</param>
        /// <returns></returns>
        private List<string> GetSetTargets(string serviceName, bool filterBlack = true)
        {
            if (!_discoveries.TryGetValue(serviceName, out IEndpointDiscovery discovery))
                return null;

            var targets = discovery.FindServiceEndpoints(filterBlack);
            if ((targets?.Count ?? 0) <= 0)
            {
                // 如果consul 取不到 暂时直接使用本地缓存的连接（注册中心数据清空的情况--异常）
                _targets.TryGetValue(serviceName, out List<string> existTargets);
                return existTargets;
            }

            var newTargets = targets.Select(oo => oo.target).ToList();
            _targets.AddOrUpdate(serviceName, newTargets, (key, value) => newTargets);
            return newTargets;
        }
#endregion
    }
}